"""Классы-исключения"""


class ServerError(Exception):
    """Исключение - ошибка сервера."""
    def __init__(self, text):
        self.text = text

    def __str__(self):
        return self.text


class NonDictError(Exception):
    """Исключение - аргумент функции не словарь."""
    def __str__(self):
        return 'Аргумент функции должен быть словарём.'


class NonStrError(Exception):
    """Исключение - аргумент функции не строка."""
    def __str__(self):
        return 'Аргумент функции должен быть строкой.'


class NonBytesError(Exception):
    """Исключение - аргумент функции не bytes."""
    def __str__(self):
        return 'Аргумент функции должен быть bytes.'


class ReqFieldMissingError(Exception):
    """Ошибка - отсутствует обязательное поле в принятом словаре."""
    def __init__(self, missing_field):
        self.missing_field = missing_field

    def __str__(self):
        return f'В принятом словаре отсутствует \
                 обязательное поле {self.missing_field}.'


class AccountNameNotUniq(Exception):
    """Ошибка - не уникальное имя пользователя."""
    def __str__(self):
        return 'Имя пользователя уже используется.'



""" Программа сервера для получения приветствия от клиента
и отправки ответа """
import sys
import os
import socket
import logging
import binascii
import hmac
from select import select
from threading import Thread
from configparser import ConfigParser
from PyQt5.QtWidgets import QApplication, QMessageBox
from server_gui import MainWindow, LoginHistoryWindow, \
                        MessageHistoryWindow, ConfigWindow, \
                        gui_create_model, create_stat_login, \
                        create_stat_message
sys.path.append('../')
import common.settings as cmnset
import common.utils as cmnutils
from common.decors import login_required
from metaclasses import ServerMaker
from server_db import ServerDB


# Инициализация серверного логера
SERVER_LOGGER = logging.getLogger('server')


class PortVerifi:
    """Класс проверки правильности введенного номера порта."""
    def __set__(self, instance, value):
        if value < 1024 and value > 65535:
            SERVER_LOGGER.error('Номер порта за пределами \
                                    диапазона 1024-65535')
            print('Ошибка: номер порта за пределами \
                            диапазона 1024-65535')
            sys.exit(1)
        else:
            instance.__dict__[self.name] = value

    def __set_name__(self, owner, name):
        # owner - <class '__main__.Server'>
        # name - port
        self.name = name


class MyServer(Thread, metaclass=ServerMaker):
    """Класс создания серверного приложения."""
    port = PortVerifi()

    def __init__(self, listen_address, listen_port, database):
        # Параметры подключения
        self.addr = listen_address
        self.port = listen_port
        # База данных сервера
        self.database = database
        # список подключенных клиентов
        self.clients = []
        # список активных пользователей и их сообщения
        self.messages = dict()
        # messages = {account_name:{'socket': client, 'message': message}}

        # Конструктор предка
        super().__init__()

    def init_socket(self):
        SERVER_LOGGER.info('Готовим сокет')
        # готовим сокет
        SERV_SOCK = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        # в таком варианте не проходит проверка в метаклассе
        # SERV_SOCK = socket(AF_INET, SOCK_STREAM)
        SERV_SOCK.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        SERV_SOCK.bind((self.addr, self.port))
        SERV_SOCK.settimeout(1)
        # Начинаем слушать сокет.
        self.SERV_SOCK = SERV_SOCK
        self.SERV_SOCK.listen(cmnset.MAX_CONNECTIONS)

    def run(self):
        # Инициализация Сокета
        self.init_socket()

        while True:  # Основной цикл программы сервера
            try:
                client, client_address = self.SERV_SOCK.accept()
            except OSError:
                pass
            else:
                print(f"Получен запрос на соединение от \
                    {str(client_address)}")
                self.clients.append(client)
            finally:
                wait = 0
                clients_read = []
                clients_write = []
                try:
                    clients_read, clients_write, errors = select(self.clients,
                                                                 self.clients,
                                                                 [],
                                                                 wait)
                except Exception:
                    pass

                for client_read in clients_read:
                    try:
                        message_from_client = cmnutils.get_message(client_read)
                        SERVER_LOGGER.debug("Получено сообщение от клиента \
                                                %s: %s",
                                            client_read,
                                            message_from_client)
                        print("Получено сообщение от клиента:\n",
                              client, message_from_client)
                    except Exception:
                        SERVER_LOGGER.info(f'Клиент {client_read} отключился \
                                                от сервера.')
                        print(f'Клиент {client_read} отключился от сервера.')
                        self.clients.remove(client_read)
                        # ищем имя клиента
                        user_name_delete = ''
                        for key in self.messages.keys():
                            if self.messages[key]['socket'] == client_read:
                                user_name_delete = key
                        # удаляем клиента из активных
                        if user_name_delete:
                            del self.messages[user_name_delete]
                            self.database.user_logout(user_name_delete)
                    else:
                        self.process_client_message(
                            message_from_client,
                            client_read)

                for sender in self.messages:
                    # print('sender:', sender)
                    # messages = {account_name:{'socket': client,
                    #                           'message': message}}
                    if 'message' in self.messages[sender] \
                            and self.messages[sender]['message']['text']:
                        if self.messages[sender]['message']['destination'] in self.messages.keys():
                            try:
                                recipient = self.messages[sender]['message']['destination']
                                cmnutils.send_message(self.messages[recipient]['socket'],
                                                      self.messages[sender]['message'])
                                self.messages[sender]['message']['text'] = ''
                                SERVER_LOGGER.debug(
                                    "Сообщение отправлено клиенту %s",
                                    self.messages[sender]['message']['destination'])
                                cmnutils.send_message(
                                    self.messages[sender]['socket'],
                                    {'response': 200,
                                     'text': 'Сообщение отправлено'})
                            except Exception:
                                SERVER_LOGGER.error(
                                    "Не удается отправить сообщение клиенту %s",
                                    self.messages[sender]['message']['destination'])
                                print(f"Не удается отправить сообщение клиенту \
                                        {self.messages[sender]['message']['destination']}")

                                self.clients.remove(self.messages[recipient]['socket'])
                                self.messages[sender]['message']['text'] = ''
                                cmnutils.send_message(
                                    self.messages[sender]['socket'],
                                    {'response': 400,
                                     'text': 'Сообщение не отправлено'})
                        else:
                            cmnutils.send_message(
                                self.messages[sender]['socket'],
                                {'response': 201,
                                 'text': 'Получатель с таким именем не активен'})
                            self.messages[sender]['message']['text'] = ''

    @login_required
    def process_client_message(self, message, client):
        """
        Функция обработки сообщений от клиентов. Производит авторизацию приложения клиента,
        регистрацию клиента по логину и паролю. Пароль клиента хранится в виде хеша в БД.
        Производит отправку сообщений между пользователями и сохранение переписки.

        """
        SERVER_LOGGER.info('проверка сообщения от клента')
        if 'action' in message \
                and 'time' in message \
                and 'user' in message:

            # авторизация и регистрация пользователя
            if message['action'] == 'presence':
                # если такого имени еще не было,
                # т.е. пользователь не активен или не зарегистрирован
                login = message['user']['account_name']
                if login not in self.messages.keys():
                    # проводим авторизацию
                    response = {'response': 511}
                    random_str = binascii.hexlify(os.urandom(64))
                    # В словарь байты нельзя, декодируем
                    # (json.dumps -> TypeError)
                    response['data'] = random_str.decode('ascii')
                    # формируем секретный ключ для авторизации на сервере
                    line_bytes = open('../common/utils.py', 'rb').readlines()
                    secret_key = b""
                    for el in line_bytes:
                        secret_key += el
                    # вычисляем хэш ключай
                    hash = hmac.new(secret_key, random_str, 'MD5')
                    digest = hash.digest()
                    try:  # Обмен с клиентом
                        cmnutils.send_message(client, response)
                        ans = cmnutils.get_message(client)
                    except OSError as err:
                        SERVER_LOGGER.debug('Error in auth, data:',
                                            exc_info=err)
                        self.socket.close()
                        return

                    client_digest = binascii.a2b_base64(ans['data'])
                    # Если ответ клиента корректный, то сохраняем его
                    # в список пользователей.
                    if 'response' in ans and ans['response'] == 511 and \
                            hmac.compare_digest(digest, client_digest):

                        # messages = {account_name:{'socket': client,
                        #                           'message': message}}
                        login = message['user']['account_name']
                        self.messages[login] = {'socket': client}

                        # добавляем клиента в список активных
                        username = message['user']['account_name']
                        password = message['user']['password']
                        pubkey = message['user']['public_key']
                        client_ip, client_port = client.getpeername()
                        login_success = self.database \
                                            .user_login(username,
                                                        password,
                                                        pubkey,
                                                        client_ip,
                                                        client_port)
                        if login_success:
                            response = {'response': 200}
                        else:
                            response = {'response': 400,
                                        'text': 'Не верное имя пользователя \
                                        или пароль'}
                        # отправляем сообщение клиенту
                        cmnutils.send_message(client, response)
                    else:
                        response = {'response': 400,
                                    'text': 'Не успешная авторизация'}
                        # отправляем сообщение клиенту
                        cmnutils.send_message(client, response)
                else:
                    SERVER_LOGGER.error('Имя пользователя %s уже занято',
                                        message['user']['account_name'])
                    response = {'response': 400,
                                'text': 'Имя пользователя уже занято'}
                    cmnutils.send_message(client, response)
                    return response

            # Отправка сообщения другому пользователю
            if message['action'] == 'message' \
                    and 'text' in message \
                    and 'destination' in message \
                    and message['destination']:

                # сохраняем сообщение в базе данных
                self.database.process_message(message)

                # если получатель активен, то запоминаем сообщение для отправки
                login = message['user']['account_name']
                if client == self.messages[login]['socket']:
                    # messages = {account_name:{'socket': client,
                    #                           'message': message}}
                    user_name = message['user']['account_name']
                    self.messages[user_name]['message'] = message
                    response = {'response': 200,
                                'text': 'сообщение поставлено \
                                в очередь отправки'}
                    # cmnutils.send_message(client, response)
                    return response
                else:
                    SERVER_LOGGER.error(f"Пользователь \
                                        {message['user']['account_name']} \
                                        с сокетом {client} не активен")
                    print(f"Пользователь {message['user']['account_name']} \
                            с сокетом {client} не активен")
                    response = {'response': 400,
                                'text': 'Пользователь не активен'}
                    cmnutils.send_message(client, response)
                    return response

            # Запрос переписки пользователя
            if message['action'] == 'get_messages' \
                    and message['user']['account_name']:
                # логин пользователя, для которого выгружаем переписку
                user_login = message['user']['account_name']
                # формируем сообщение для отправки
                response = {'response': 200}
                response['text'] = self.database.get_messages(user_login)
                print('!!!! get_messages:', response)
                # отправляем сообщение
                cmnutils.send_message(client, response)
                return(response)

            # Запрос списка контактов для пользователя
            if message['action'] == 'get_contacts':
                response = {'response': 200}
                user_login = message['user']['account_name']
                response['text'] = self.database.get_contacts(user_login)
                cmnutils.send_message(client, response)
                return(response)

            # Запрос списка всех пользователей
            if message['action'] == 'get_userlist':
                response = {'response': 200}
                response['text'] = self.database.get_userlist()
                print(response)
                cmnutils.send_message(client, response)
                return(response)

            # Добавление контакта
            if message['action'] == 'add_contact' \
                    and 'destination' in message \
                    and message['destination']:

                user_login = message['user']['account_name']
                contact = message['destination']

                if self.database.add_contact(user_login, contact) == 1:
                    response = {'response': 200,
                                'text': 'пользователь добавлен в контакты'}
                elif self.database.add_contact(user_login, contact) == 2:
                    response = {'response': 200,
                                'text': 'такой контакт уже существует'}
                elif self.database.add_contact(user_login, contact) == 0:
                    response = {'response': 400,
                                'text': 'такого пользователя нет'}
                cmnutils.send_message(client, response)
                return response

            # Удаление контакта
            if message['action'] == 'del_contact' \
                    and 'destination' in message \
                    and message['destination']:

                user_login = message['user']['account_name']
                contact = message['destination']

                if self.database.del_contact(user_login, contact) == 1:
                    response = {'response': 200,
                                'text': 'пользователь удален из контактов'}
                elif self.database.del_contact(user_login, contact) == 2:
                    response = {'response': 400,
                                'text': 'такого контакта не существует'}
                elif self.database.del_contact(user_login, contact) == 0:
                    response = {'response': 400,
                                'text': 'такого пользователя нет'}
                cmnutils.send_message(client, response)
                return response

            # выход клиента с сервера
            if message['action'] == 'exit':
                self.clients.remove(client)
                client.close()
                # удаляем клиента из активных
                del self.messages[message['user']['account_name']]
                self.database.user_logout(message['user']['account_name'])

            SERVER_LOGGER.debug('сообщение от клента правильное')
            return {'response': 200}
        else:
            SERVER_LOGGER.error('сообщение от клента не правильное')
            return {'response': 400,
                    'text': 'bad request'}


def print_help():
    print(50*'=')
    print('Поддерживаемые комманды:')
    print('gui       - графический интерфейс сервера')
    print('users     - список известных пользователей')
    print('connected - список подключённых пользователей')
    print('loghist   - история входов пользователя')
    print('exit      - завершение работы сервера.')
    print('help      - вывод справки по поддерживаемым командам')
    print(50*'=')


def main():
    SERVER_LOGGER.info('Определяем параметры сервера')
    config = ConfigParser()

    dir_path = os.path.dirname(os.path.abspath(__file__))
    config.read(os.path.join(dir_path, 'server.ini'),
                encoding='utf-8')
    DEFAULT_PORT = int(config['SETTINGS']['default_port'])
    DEFAULT_ADDR = config['SETTINGS']['listen_address']

    try:
        if '-p' in sys.argv:
            listen_port = int(sys.argv[sys.argv.index('-p') + 1])
        else:
            SERVER_LOGGER.debug('Используется порт по умолчанию %d',
                                DEFAULT_PORT)
            listen_port = DEFAULT_PORT
    except IndexError:
        SERVER_LOGGER.error('После параметра -\'p\' необходимо \
                            указать номер порта.')
        sys.exit(1)

    try:
        if '-a' in sys.argv:
            listen_address = int(sys.argv[sys.argv.index('-a') + 1])
        else:
            SERVER_LOGGER.info('Используется адрес по умолчанию')
            listen_address = DEFAULT_ADDR
    except IndexError:
        SERVER_LOGGER.error('После параметра -\'а\' необходимо \
                            указать адрес для прослушивания сервером.')
        sys.exit(1)

    # Инициализация базы данных
    database = ServerDB(
                    os.path.join(
                        config['SETTINGS']['database_path'],
                        config['SETTINGS']['database_file'])
                    )

    # Создание экземпляра класса - сервера.
    server = MyServer(listen_address, listen_port, database)
    server.daemon = True
    server.start()

    def server_gui():
        """Создаёт графическое окружение для сервера."""
        server_app = QApplication(sys.argv)
        main_window = MainWindow()

        # Функция, обновляющая список подключённых клиентов
        def list_update():
            main_window.active_clients_table \
                .setModel(gui_create_model(database))
            main_window.active_clients_table \
                .resizeColumnsToContents()
            main_window.active_clients_table \
                .resizeRowsToContents()

        # Инициализируем параметры окна
        main_window.statusBar().showMessage('Server Working')
        list_update()

        # Функция, создающая окно со статистикой клиентов
        def login_statistics():
            global stat_window
            stat_window = LoginHistoryWindow()
            stat_window.history_table.setModel(create_stat_login(database))
            stat_window.history_table.resizeColumnsToContents()
            stat_window.history_table.resizeRowsToContents()
            stat_window.show()

        def message_statistics():
            global stat_window
            stat_window = MessageHistoryWindow()
            stat_window.history_table.setModel(create_stat_message(database))
            stat_window.history_table.resizeColumnsToContents()
            stat_window.history_table.resizeRowsToContents()
            stat_window.show()

        # Функция создающяя окно с настройками сервера.
        def server_config():
            global config_window
            # Создаём окно и заносим в него текущие параметры
            config_window = ConfigWindow()
            config_window.db_path.insert(config['SETTINGS']['database_path'])
            config_window.db_file.insert(config['SETTINGS']['database_file'])
            config_window.port.insert(config['SETTINGS']['default_port'])
            config_window.ip.insert(config['SETTINGS']['listen_address'])
            config_window.save_btn.clicked.connect(save_server_config)

        # Функция сохранения настроек
        def save_server_config():
            global config_window
            message_box = QMessageBox()
            config['SETTINGS']['database_path'] = config_window.db_path.text()
            config['SETTINGS']['database_file'] = config_window.db_file.text()

            try:
                port = int(config_window.port.text())
            except ValueError:
                message_box.warning(config_window,
                                    'Ошибка',
                                    'Порт должен быть числом')
            else:
                config['SETTINGS']['listen_address'] = config_window.ip.text()
                if 1023 < port < 65536:
                    config['SETTINGS']['default_port'] = str(port)
                    with open('server.ini', 'w', encoding='utf-8') as conf:
                        config.write(conf)
                        message_box.information(
                            config_window,
                            'OK',
                            'Настройки успешно сохранены!')
                else:
                    message_box.warning(
                        config_window,
                        'Ошибка',
                        'Порт должен быть от 1024 до 65536')

        # Связываем кнопки с процедурами
        main_window.refresh_button \
            .triggered \
            .connect(list_update)
        main_window.login_history_button \
            .triggered \
            .connect(login_statistics)
        main_window.message_history_button \
            .triggered \
            .connect(message_statistics)
        main_window.config_btn \
            .triggered \
            .connect(server_config)

        # Таймер, обновляющий список клиентов 1 раз в секунду
        # timer = QTimer()
        # timer.timeout.connect(list_update)
        # timer.start(1000)

        # Запускаем GUI
        server_app.exec_()

    # консольная часть
    print('Сервер запущен!')
    # Печатаем справку:
    print_help()

    # Основной цикл сервера:
    while True:
        command = input('Введите команду: ')
        if command == 'help':
            print_help()
        elif command == 'exit':
            break
        elif command == 'gui':
            server_gui()
        elif command == 'users':
            for user in sorted(database.users_list()):
                print(f'Пользователь {user[0]}, \
                      последний вход: {user[1]}')
        elif command == 'connected':
            for user in sorted(database.active_users_list()):
                print(f'Пользователь {user[0]}, \
                      подключен: {user[1]}:{user[2]}, \
                      время установки соединения: {user[3]}')
        elif command == 'loghist':
            name = input('Введите имя пользователя для просмотра истории. '
                         'Для вывода всей истории, просто нажмите Enter: ')
            for user in sorted(database.login_history(name)):
                print(f'Пользователь: \
                        {user[0]} время входа: {user[1]}. \
                        Вход с: {user[2]}:{user[3]}')
        else:
            print('Команда не распознана.')


if __name__ == '__main__':
    main()



import os
import datetime
from sqlalchemy import create_engine, Column, Integer, \
                       String, DateTime, ForeignKey
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker
from pprint import pprint


class ServerDB:
    """Класс создания базы данных сервера."""
    Base = declarative_base()

    class AllUsers(Base):
        """
        Класс создания таблицы пользователей.
        Содержит поля:
        id - порядковый номер пользователя,
        login - имя пользователя,
        passwd_hash - пароль в виде хэша,
        pubkey - публичный ключ для авторизации на сервере,
        ip - адрес, с которого клиент подключился к серверу,
        port - порт подключения клиента
        is_active - признак активности клиента
        last_conn - дата и время последнего подключения
        """
        __tablename__ = 'users'
        id = Column(Integer, primary_key=True)
        login = Column(String, unique=True)
        passwd_hash = Column(String, default=None)
        pubkey = Column(String, default=None)
        ip = Column(String, default='0.0.0.0')
        port = Column(Integer, default=0)
        is_active = Column(Integer, default=0)
        last_conn = Column(DateTime)

        def __init__(self, login, passwd_hash, pubkey, ip, port):
            self.login = login
            self.passwd_hash = passwd_hash
            self.pubkey = pubkey
            self.ip = ip
            self.port = port
            self.is_active = 1
            self.last_conn = datetime.datetime.now()

    class UsersContacts(Base):
        """
        Класс создания таблицы контактов.
        Содержит поля:
        id - порядковый номер записи,
        user_id - id пользователя для которого сохранен контакт,
        contact - id пользователя с которым сохранен контакт
        """
        __tablename__ = 'contacts'
        id = Column(Integer, primary_key=True)
        user_id = Column(Integer, ForeignKey('users.id'))
        contact = Column(Integer, ForeignKey('users.id'))

        def __init__(self, user_id, contact):
            self.user_id = user_id
            self.contact = contact

    class LoginHistory(Base):
        """
        Класс создания таблицы истории подключения к серверу.
        Таблица накопительная. Содержит столбцы:
        id - порядковый номер записи,
        user_id - id подключенного пользователя,
        ip - адрес подключения пользователя,
        port - порт подключения пользователя,
        last_conn - дата и время подключения
        """
        __tablename__ = 'login_history'
        id = Column(Integer, primary_key=True)
        user_id = Column(Integer, ForeignKey('users.id'))
        ip = Column(String)
        port = Column(Integer)
        last_conn = Column(DateTime)

        def __init__(self, user_id, ip, port, last_conn):
            self.user_id = user_id
            self.ip = ip
            self.port = port
            self.last_conn = last_conn

    class MessageHistory(Base):
        """
        Класс создания таблицы статистики переписки.
        Содержит столбцы:
        id - порядковый номер записи,
        user_id - id пользователя,
        cnt_sent - количество отправленных сообщений за всё время,
        cnt_receive - количество принятых сообщений за всё время
        """
        __tablename__ = 'message_history'
        id = Column(Integer, primary_key=True)
        user_id = Column(Integer, ForeignKey('users.id'))
        cnt_sent = Column(Integer)
        cnt_receive = Column(Integer)

        def __init__(self, user_id):
            self.user_id = user_id
            self.cnt_sent = 0
            self.cnt_receive = 0

    class Messages(Base):
        """
        Класс создания таблицы истории переписки.
        Таблица накопительная. Содержит столбцы:
        id - порядковый номер записи,
        sender - id пользователя отправителя,
        destination - id пользователя получателя,
        text - текст сообщения,
        date - дата и время сохранения сообщения в БД
        """
        __tablename__ = 'messages'
        id = Column(Integer, primary_key=True)
        sender = Column(Integer, ForeignKey('users.id'))
        destination = Column(Integer, ForeignKey('users.id'))
        text = Column(String)
        date = Column(DateTime)

        def __init__(self, sender, destination, text, date):
            self.sender = sender
            self.destination = destination
            self.text = text
            self.date = date

    def __init__(self, path):
        # Создаём движок базы данных
        self.engine = create_engine(
            f'sqlite:///{path}',
            echo=False,
            pool_recycle=7200,
            connect_args={'check_same_thread': False})

        self.Base.metadata.create_all(self.engine)
        Session = sessionmaker(bind=self.engine)
        self.session = Session()

        # Если в таблице активных пользователей есть записи,
        # то их необходимо удалить
        # Когда устанавливаем соединение, очищаем таблицу
        # активных пользователей
        # self.session.query(self.ActiveUsers).delete()
        self.session.query(self.AllUsers).update({'is_active': 0})
        self.session.commit()

    def user_login(self, username, password, pubkey, ip_address, port):
        """
        Функция выполняется при входе пользователя,
        фиксирует в базе факт входа и для первого входа создает
        пользователя в базе данных.
        """
        # Запрос в таблицу пользователей на наличие там
        # пользователя с таким именем
        result = self.session.query(self.AllUsers) \
                             .filter_by(login=username)

        # Если имя пользователя уже присутствует в таблице,
        # проверяем пароль и обновляем время последнего входа
        if result.count():
            user = result.first()
            if password == user.passwd_hash:
                user.ip = ip_address
                user.port = port
                user.is_active = 1
                user.last_conn = datetime.datetime.now()
                user.pubkey = pubkey
            else:
                return False  # пароль не совпадает
        else:  # Если нет, то создаём нового пользователя
            user = self.AllUsers(username,
                                 password,
                                 pubkey,
                                 ip_address,
                                 port)
            self.session.add(user)
            # Коммит здесь нужен, чтобы в db записался ID
            self.session.commit()
            user_in_message_history = self.MessageHistory(user.id)
            self.session.add(user_in_message_history)

        # и сохранить в историю входов
        # Создаем экземпляр класса self.LoginHistory,
        # через который передаем данные в таблицу
        history = self.LoginHistory(user.id,
                                    ip_address,
                                    port,
                                    datetime.datetime.now())
        self.session.add(history)
        self.session.commit()
        return True  # успешная регистрация

    def user_logout(self, username):
        """Функция фиксирует отключение пользователя."""
        # Запрашиваем пользователя, что покидает нас
        # получаем запись из таблицы AllUsers
        # print(f'user_logout({username})')
        user = self.session.query(self.AllUsers) \
                           .filter_by(login=username) \
                           .first()
        user.is_active = 0
        # Применяем изменения
        self.session.commit()

    def active_users_list(self):
        """Функция возвращает список активных пользователей"""
        # Запрашиваем соединение таблиц и собираем тюплы
        # имя, адрес, порт, время.
        query = self.session.query(self.AllUsers.login,
                                   self.AllUsers.ip,
                                   self.AllUsers.port,
                                   # self.AllUsers.is_active,
                                   self.AllUsers.last_conn) \
                            .filter_by(is_active=1)
        # Возвращаем список тюплов
        return query.all()

    def login_history(self, username=None):
        """
        Функция возвращает историю входов по пользователю
        или по всем пользователям.
        """
        # Запрашиваем историю входа
        query = self.session.query(self.AllUsers.login,
                                   self.LoginHistory.last_conn,
                                   self.LoginHistory.ip,
                                   self.LoginHistory.port
                                   ).join(self.AllUsers)
        # Если было указано имя пользователя, то фильтруем по нему
        if username:
            query = query.filter(self.AllUsers.login == username)
        return query.all()

    def message_history(self):
        """Функция возвращает статистику переданных и полученных сообщений."""
        query = self.session.query(
            self.AllUsers.login,
            self.AllUsers.last_conn,
            self.MessageHistory.cnt_sent,
            self.MessageHistory.cnt_receive
        ).join(self.AllUsers)
        # Возвращаем список кортежей
        return query.all()

    def process_message(self, message):
        """
        Функция фиксирует передачу сообщения
        и делает соответствующие отметки в БД.
        """
        sender = message['user']['account_name']
        destination = message['destination']
        text = message['text']
        # Получаем ID отправителя и получателя
        sender = self.session.query(self.AllUsers) \
                             .filter_by(login=sender) \
                             .first().id
        destination = self.session.query(self.AllUsers) \
                                  .filter_by(login=destination) \
                                  .first().id
        # Запрашиваем строки из истории и увеличиваем счётчики
        sender_row = self.session.query(self.MessageHistory) \
                                 .filter_by(user_id=sender) \
                                 .first()
        sender_row.cnt_sent += 1
        recipient_row = self.session.query(self.MessageHistory) \
                                    .filter_by(user_id=destination) \
                                    .first()
        recipient_row.cnt_receive += 1
        # сохраняем сообщение
        row = self.Messages(sender, destination, text, datetime.datetime.now())
        self.session.add(row)
        self.session.commit()

    def get_messages(self, username):
        """Функция возвращает всю переписку пользователя."""
        # Получаем ID пользователей
        user = self.session.query(self.AllUsers) \
                                .filter_by(login=username) \
                                .first()
        # список исходящих сообщений для sender
        req_1 = self.session.query(self.Messages) \
                            .filter_by(sender=user.id)
        # список входящих сообщений для sender
        req_2 = self.session.query(self.Messages) \
                            .filter_by(destination=user.id)
        # sender, destination, text, date
        messages = []
        for el in req_1.all():
            # 2022-06-26 15:02:52
            messages.append(('out',
                             el.destination,
                             el.text,
                             el.date.strftime('%Y-%m-%d %H:%M:%S')))

        for el in req_2.all():
            messages.append(('in',
                             el.sender,
                             el.text,
                             el.date.strftime('%Y-%m-%d %H:%M:%S')))
        # print(messages)
        return messages

    def add_contact(self, login, contact):
        """Функция добавляет контакт для пользователя."""
        # Получаем ID пользователей
        user = self.session.query(self.AllUsers) \
                           .filter_by(login=login) \
                           .first()
        contact = self.session.query(self.AllUsers) \
                              .filter_by(login=contact) \
                              .first()
        # Проверяем что не дубль и что контакт может
        # существовать (полю пользователь мы доверяем)
        if not contact:
            return 0  # такого пользователя нет в системе

        if self.session.query(self.UsersContacts) \
                       .filter_by(user_id=user.id,
                                  contact=contact.id) \
                       .count():
            return 2  # такой контакт уже есть
        # Создаём объект и заносим его в базу
        contact_row = self.UsersContacts(user.id, contact.id)
        self.session.add(contact_row)
        self.session.commit()
        return 1  # контакт добавлен

    def del_contact(self, login, contact):
        """Функция удаляет контакт из базы данных."""
        # Получаем ID пользователей
        user = self.session.query(self.AllUsers) \
                           .filter_by(login=login) \
                           .first()
        contact = self.session.query(self.AllUsers) \
                              .filter_by(login=contact) \
                              .first()

        # Проверяем что контакт может существовать
        # (полю пользователь мы доверяем)
        if not contact:
            return 0  # такого пользователя нет в системе

        if not self.session.query(self.UsersContacts) \
                           .filter_by(user_id=user.id,
                                      contact=contact.id) \
                           .count():
            return 2  # такой контакт не существует

        # Удаляем требуемое
        self.session.query(self.UsersContacts) \
                    .filter(self.UsersContacts.user_id == user.id,
                            self.UsersContacts.contact == contact.id) \
                    .delete()
        self.session.commit()
        return 1

    def get_contacts(self, login):
        """Функция возвращает список контактов пользователя."""
        # Запрашиваем указанного пользователя
        user = self.session.query(self.AllUsers) \
                           .filter_by(login=login) \
                           .first()

        # Запрашиваем его список контактов
        query = self.session.query(self.UsersContacts,
                                   self.AllUsers.login) \
                            .filter_by(user_id=user.id) \
                            .join(self.AllUsers,
                                  self.UsersContacts.contact
                                  == self.AllUsers.id)

        # выбираем только имена пользователей и возвращаем их.
        return [contact[1] for contact in query.all()]

    def get_userlist(self):
        """Функция возвращает список известных пользователей."""
        query = self.session.query(
            self.AllUsers.login,
            self.AllUsers.last_conn,
        )
        # Возвращаем список тюплов
        # return query.all()
        return [contact[0] for contact in query.all()]

    def get_hash(self, login):
        """Метод получения хэша пароля пользователя."""
        user = self.session.query(self.AllUsers) \
                           .filter_by(login=login) \
                           .first()
        return user.passwd_hash

    def get_pubkey(self, login):
        """Метод получения публичного ключа пользователя."""
        user = self.session.query(self.AllUsers) \
                           .filter_by(login=login) \
                           .first()
        return user.pubkey


if __name__ == '__main__':
    path = os.path.dirname(os.path.abspath(__file__))
    db = ServerDB(os.path.join(path, 'server_db.db3'))

    print(50*'=')
    print('login user test_1')
    db.user_login('test_1', b'password_test_1',
                  b'public key test_1', '192.168.1.4', 8888)
    print('login user test_2')
    db.user_login('test_2', b'password_test_2',
                  b'public key test_2', '192.168.1.5', 7777)
    print('login user test_3')
    db.user_login('test_3', b'password_test_3',
                  b'public key test_3', '192.168.1.6', 7778)

    # выводим список активных пользователей
    print(50*'=')
    print('active users list:')
    pprint(db.active_users_list())

    # выполянем 'отключение' пользователя
    print(50*'=')
    print('logout user test_1')
    db.user_logout('test_1')

    # выводим список всех пользователей
    print(50*'=')
    print('all users list:')
    pprint(db.get_userlist())

    # выводим список активных пользователей
    print(50*'=')
    print('active users list:')
    pprint(db.active_users_list())

    print(50*'=')
    print('logout user test_2')
    db.user_logout('test_2')

    print(50*'=')
    print('all users list:')
    pprint(db.get_userlist())

    print(50*'=')
    print('active users list:')
    pprint(db.active_users_list())

    print(50*'=')
    print('login user test_1 with wrong password')
    if db.user_login('test_1', b'password_test1',
                     b'public key test_1', '192.168.1.4', 8888):
        print('success')
    else:
        print('not success')

    # запрашиваем историю входов по пользователю
    print(50*'=')
    print('login history for user test_1:')
    pprint(db.login_history('test_1'))

    print(50*'=')
    print('add contacts test_2 - test_1')
    db.add_contact('test_2', 'test_1')
    print('add contacts test_1 - test_3')
    db.add_contact('test_1', 'test_3')
    print('add contacts test_1 - test_6')
    db.add_contact('test_1', 'test_6')

    print(50*'=')
    print('print contacts user test_1')
    pprint(db.get_contacts('test_1'))
    print()
    print('print contacts user test_2')
    pprint(db.get_contacts('test_2'))

    print(50*'=')
    print('remove contacts test_1 - test_3')
    db.del_contact('test_1', 'test_3')
    print('print contacts user test_1')
    pprint(db.get_contacts('test_1'))

    print(50*'=')
    print('send message from test_1 to test_3')
    message = {'action': 'message',
               'time': datetime.datetime.now(),
               'user': {'account_name': 'test_1'},
               'destination': 'test_3',
               'text': 'some message from test_1 to test_3'}
    db.process_message(message)

    print(50*'=')
    print('get messages test_1 - test_3')
    db.get_messages('test_1')

    print(50*'=')
    print('get messages test_3 - test_1')
    db.get_messages('test_3')



import os
import sys
import unittest
sys.path.insert(0, os.path.join(os.getcwd(), '..'))
from client import create_presence, process_ans


class TestCreatePresence(unittest.TestCase):
    def test_equal(self):
        self.assertEqual(create_presence()['action'], 'presence')
        self.assertEqual(create_presence()['user']['account_name'], 'Guest')
        self.assertEqual(
            create_presence('User1')['user']['account_name'], 'User1')

    def test_not_equal(self):
        self.assertNotEqual(
            create_presence('User1')['user']['account_name'], 'Guest')

    def test_isinstance(self):
        self.assertIsInstance(create_presence(), dict)
        self.assertIsInstance(create_presence()['action'], str)
        self.assertIsInstance(create_presence()['time'], float)
        self.assertIsInstance(create_presence()['user'], dict)
        self.assertIsInstance(create_presence()['user']['account_name'], str)


class TestProcessAns(unittest.TestCase):
    def test_isinstance(self):
        message = {'response': 200}
        self.assertIsInstance(process_ans(message), dict)

    def test_equal(self):
        message_ok = {'response': 200}
        message_er = {'response': 201, 'error': 'bad request'}
        self.assertEqual(process_ans(message_ok), {'200': 'ok'})
        self.assertEqual(process_ans(message_er), '400: bad request')

    def test_not_equal(self):
        message_ok = {'response': 200}
        message_er = {'response': 201, 'error': 'bad request'}
        self.assertNotEqual(process_ans(message_ok), {'200': ''})
        self.assertNotEqual(process_ans(message_er), 'bad request')

    def test_raises(self):
        message_ok = {'rsponse': 200}
        self.assertRaises(ValueError, process_ans, message_ok)


if __name__ == '__main__':
    unittest.main()



import unittest
import json
import sys
import os
from socket import socket, AF_INET, SOCK_STREAM, SOL_SOCKET, SO_REUSEADDR

sys.path.insert(0, os.path.join(os.getcwd(), '..'))
from common.utils import get_message, send_message
from common.settings import ENCODING, MAX_CONNECTIONS, \
                            DEFAULT_ADDRESS, DEFAULT_PORT, \
                            MAX_PACKAGE_LENGTH


class TestUtils(unittest.TestCase):
    test_message = {
        'action': 'presence',
        'time': 11.11,
        'user': {
            'account_name': 'Guest',
        },
    }

    response_ok = {'response': 200,
                   'alert': 'соединение успешно'}
    response_err = {'response': 400,
                    'error': 'ошибка соединения'}

    server_socket = None
    client_socket = None

    def setUp(self) -> None:
        # создаем тестовый сокет для сервера
        self.server_socket = socket(AF_INET, SOCK_STREAM)
        self.server_socket.setsockopt(SOL_SOCKET, SO_REUSEADDR, 1)
        self.server_socket.bind((DEFAULT_ADDRESS, DEFAULT_PORT))
        self.server_socket.listen(MAX_CONNECTIONS)
        # создаем тестовый сокет для клиента
        self.client_socket = socket(AF_INET, SOCK_STREAM)
        self.client_socket.connect((DEFAULT_ADDRESS, DEFAULT_PORT))
        self.client, self.client_address = self.server_socket.accept()

    def tearDown(self) -> None:
        self.client.close()
        self.client_socket.close()
        self.server_socket.close()

    def test_send_wrong_message_from_client(self):
        self.assertRaises(TypeError,
                          send_message,
                          self.client_socket,
                          'not dict')

    def test_send_message_client_server(self):
        send_message(self.client_socket, self.test_message)
        test_response = self.client.recv(MAX_PACKAGE_LENGTH)
        test_response = json.loads(test_response.decode(ENCODING))
        self.client.close()
        self.assertEqual(self.test_message, test_response)

    def test_get_message_200(self):
        message = json.dumps(self.response_ok)
        self.client.send(message.encode(ENCODING))
        self.client.close()

        response = get_message(self.client_socket)
        self.assertEqual(self.response_ok, response)

    def test_get_message_400(self):
        message = json.dumps(self.response_err)
        self.client.send(message.encode(ENCODING))
        self.client.close()

        response = get_message(self.client_socket)
        self.assertEqual(self.response_err, response)


if __name__ == '__main__':
    unittest.main()



import unittest
import json
import sys
import os
sys.path.insert(0, os.path.join(os.getcwd(), '..'))
from common.utils import get_message, send_message
from common.settings import ENCODING


class TestSocket:
    def __init__(self, test_dic):
        self.test_dic = test_dic
        self.encoded_message = None
        self.received_message = None

    def send(self, message_to_send):
        json_test_message = json.dumps(self.test_dic)
        self.encoded_message = json_test_message.encode(ENCODING)
        self.received_message = message_to_send

    def recv(self, max_len):
        json_test_message = json.dumps(self.test_dic)
        return json_test_message.encode(ENCODING)


class TestUtils(unittest.TestCase):
    test_dic_send = {
        'action': 'presence',
        'time': 111.111,
        'user': {
            'account_name': 'test_user',
        },
    }

    response_ok = {'response': 200}
    response_err = {'response': 400, 'error': 'bad request'}

    def test_send_message_true(self):
        test_socket = TestSocket(self.test_dic_send)
        send_message(test_socket, self.test_dic_send)
        self.assertEqual(
            test_socket.encoded_message, test_socket.received_message)

    def test_send_message_with_error(self):
        test_socket = TestSocket(self.test_dic_send)
        send_message(test_socket, self.test_dic_send)
        self.assertRaises(TypeError, send_message, test_socket, 'wrong_dict')

    def test_get_message_ok(self):
        test_socket_ok = TestSocket(self.response_ok)
        self.assertEqual(get_message(test_socket_ok), self.response_ok)

    def test_get_message_err(self):
        test_socket_err = TestSocket(self.response_err)
        self.assertEqual(get_message(test_socket_err), self.response_err)


if __name__ == '__main__':
    unittest.main()



import sys
from PyQt5.QtWidgets import QMainWindow, QAction, qApp, QApplication, \
    QLabel, QTableView, QDialog, QPushButton, QLineEdit, QFileDialog
from PyQt5.QtGui import QStandardItemModel, QStandardItem
from PyQt5.QtCore import Qt


# GUI - Создание таблицы QModel, для отображения в окне программы.
def gui_create_model(database):
    list_users = database.active_users_list()
    # print('!!!!!! list_users =', list_users)
    list_table = QStandardItemModel()
    list_table.setHorizontalHeaderLabels(
        ['Имя Клиента', 'IP Адрес', 'Порт', 'Время подключения'])
    for row in list_users:
        user, ip, port, time = row
        user = QStandardItem(user)
        user.setEditable(False)
        ip = QStandardItem(ip)
        ip.setEditable(False)
        port = QStandardItem(str(port))
        port.setEditable(False)
        # Уберём миллисекунды из строки времени,
        # т.к. такая точность не требуется.
        time = QStandardItem(str(time.replace(microsecond=0)))
        time.setEditable(False)
        list_table.appendRow([user, ip, port, time])
    return list_table


# GUI - Функция реализующая заполнение таблицы историей сообщений.
def create_stat_message(database):
    # Список записей из базы
    hist_list = database.message_history()

    # Объект модели данных:
    list_table = QStandardItemModel()
    list_table.setHorizontalHeaderLabels(
        ['Имя Клиента', 'Дата входа',
         'Сообщений отправлено', 'Сообщений получено'])
    for row in hist_list:
        user, last_seen, sent, recvd = row

        user = QStandardItem(user)
        user.setEditable(False)

        last_seen = QStandardItem(str(last_seen.replace(microsecond=0)))
        last_seen.setEditable(False)

        sent = QStandardItem(str(sent))
        sent.setEditable(False)

        recvd = QStandardItem(str(recvd))
        recvd.setEditable(False)

        list_table.appendRow([user, last_seen, sent, recvd])
    return list_table


# GUI - Функция реализующая заполнение таблицы историй логирования
def create_stat_login(database):
    # Список записей из базы
    hist_list = database.login_history()

    # Объект модели данных:
    list_table = QStandardItemModel()
    list_table.setHorizontalHeaderLabels(
        ['Имя Клиента', 'Дата входа', 'ip адрес', 'порт'])
    for row in hist_list:
        user, last_seen, addr, port = row

        user = QStandardItem(user)
        user.setEditable(False)

        last_seen = QStandardItem(str(last_seen.replace(microsecond=0)))
        last_seen.setEditable(False)

        addr = QStandardItem(str(addr))
        addr.setEditable(False)

        port = QStandardItem(str(port))
        port.setEditable(False)

        list_table.appendRow([user, last_seen, addr, port])
    return list_table


# Класс основного окна
class MainWindow(QMainWindow):
    def __init__(self):
        super().__init__()
        self.initUI()

    def initUI(self):
        # Кнопка выхода
        self.exitAction = QAction('Выход', self)
        self.exitAction.setShortcut('Ctrl+Q')
        self.exitAction.triggered.connect(qApp.quit)

        # Кнопка обновить список клиентов
        self.refresh_button = QAction('Обновить список', self)

        # Кнопка вывести историю подключений
        self.login_history_button = QAction('История подключений', self)

        # Кнопка вывести историю подключений
        self.message_history_button = QAction('Статистика сообщений', self)

        # Кнопка настроек сервера
        self.config_btn = QAction('Настройки сервера', self)

        # Статусбар
        # dock widget
        self.statusBar()

        # Тулбар
        self.toolbar = self.addToolBar('MainBar')
        self.toolbar.addAction(self.exitAction)
        self.toolbar.addAction(self.refresh_button)
        self.toolbar.addAction(self.login_history_button)
        self.toolbar.addAction(self.message_history_button)
        self.toolbar.addAction(self.config_btn)

        # Настройки геометрии основного окна
        # Поскольку работать с динамическими размерами мы не умеем,
        # и мало времени на изучение, размер окна фиксирован.
        self.setFixedSize(800, 600)
        self.setWindowTitle('Messaging Server alpha release')

        # Надпись о том, что ниже список подключённых клиентов
        self.label = QLabel('Список подключённых клиентов:', self)
        self.label.setFixedSize(400, 15)
        self.label.move(10, 35)

        # Окно со списком подключённых клиентов.
        self.active_clients_table = QTableView(self)
        self.active_clients_table.move(10, 55)
        self.active_clients_table.setFixedSize(780, 400)

        # Последним параметром отображаем окно.
        self.show()


# Класс окна с историей подключений
class LoginHistoryWindow(QDialog):
    def __init__(self):
        super().__init__()
        self.initUI()

    def initUI(self):
        # Настройки окна:
        self.setWindowTitle('Статистика подключений')
        self.setFixedSize(600, 700)
        self.setAttribute(Qt.WA_DeleteOnClose)

        # Кнопка закрытия окна
        self.close_button = QPushButton('Закрыть', self)
        self.close_button.move(250, 650)
        self.close_button.clicked.connect(self.close)

        # Лист с собственно историей
        self.history_table = QTableView(self)
        self.history_table.move(10, 10)
        self.history_table.setFixedSize(580, 620)

        self.show()


# Класс окна с историей подключений
class MessageHistoryWindow(QDialog):
    def __init__(self):
        super().__init__()
        self.initUI()

    def initUI(self):
        # Настройки окна:
        self.setWindowTitle('Статистика сообщений')
        self.setFixedSize(600, 700)
        self.setAttribute(Qt.WA_DeleteOnClose)

        # Кнопка закрытия окна
        self.close_button = QPushButton('Закрыть', self)
        self.close_button.move(250, 650)
        self.close_button.clicked.connect(self.close)

        # Лист с собственно историей
        self.history_table = QTableView(self)
        self.history_table.move(10, 10)
        self.history_table.setFixedSize(580, 620)

        self.show()


# Класс окна настроек
class ConfigWindow(QDialog):
    def __init__(self):
        super().__init__()
        self.initUI()

    def initUI(self):
        # Настройки окна
        self.setFixedSize(365, 260)
        self.setWindowTitle('Настройки сервера')

        # Надпись о файле базы данных:
        self.db_path_label = QLabel('Путь до файла базы данных: ', self)
        self.db_path_label.move(10, 10)
        self.db_path_label.setFixedSize(240, 15)

        # Строка с путём базы
        self.db_path = QLineEdit(self)
        self.db_path.setFixedSize(250, 20)
        self.db_path.move(10, 30)
        self.db_path.setReadOnly(True)

        # Кнопка выбора пути.
        self.db_path_select = QPushButton('Обзор...', self)
        self.db_path_select.move(275, 28)

        # Функция обработчик открытия окна выбора папки
        def open_file_dialog():
            global dialog
            dialog = QFileDialog(self)
            path = dialog.getExistingDirectory()
            path = path.replace('/', '\\')
            self.db_path.clear()
            self.db_path.insert(path)

        self.db_path_select.clicked.connect(open_file_dialog)

        # Метка с именем поля файла базы данных
        self.db_file_label = QLabel('Имя файла базы данных: ', self)
        self.db_file_label.move(10, 68)
        self.db_file_label.setFixedSize(180, 15)

        # Поле для ввода имени файла
        self.db_file = QLineEdit(self)
        self.db_file.move(200, 66)
        self.db_file.setFixedSize(150, 20)

        # Метка с номером порта
        self.port_label = QLabel('Номер порта для соединений:', self)
        self.port_label.move(10, 108)
        self.port_label.setFixedSize(180, 15)

        # Поле для ввода номера порта
        self.port = QLineEdit(self)
        self.port.move(200, 108)
        self.port.setFixedSize(150, 20)

        # Метка с адресом для соединений
        self.ip_label = QLabel('С какого IP принимаем соединения:', self)
        self.ip_label.move(10, 148)
        self.ip_label.setFixedSize(180, 15)

        # Метка с напоминанием о пустом поле.
        self.ip_label_note = QLabel(
            ' оставьте это поле пустым, чтобы\n \
            принимать соединения с любых адресов.', self)
        self.ip_label_note.move(10, 168)
        self.ip_label_note.setFixedSize(500, 30)

        # Поле для ввода ip
        self.ip = QLineEdit(self)
        self.ip.move(200, 148)
        self.ip.setFixedSize(150, 20)

        # Кнопка сохранения настроек
        self.save_btn = QPushButton('Сохранить', self)
        self.save_btn.move(190, 220)
        # self.save_btn.clicked.connect(self.close)

        # Кнопка закрытия окна
        self.close_button = QPushButton('Закрыть', self)
        self.close_button.move(275, 220)
        self.close_button.clicked.connect(self.close)

        self.show()


if __name__ == '__main__':
    app = QApplication(sys.argv)
    main_window = MainWindow()
    main_window.statusBar().showMessage('Test Statusbar Message')
    test_list = QStandardItemModel(main_window)
    test_list.setHorizontalHeaderLabels(
        ['Имя Клиента', 'IP Адрес', 'Порт', 'Время подключения'])
    test_list.appendRow(
        [QStandardItem('test1'),
         QStandardItem('192.198.0.5'),
         QStandardItem('23544'),
         QStandardItem('16:20:34')])
    test_list.appendRow(
        [QStandardItem('test2'),
         QStandardItem('192.198.0.8'),
         QStandardItem('33245'),
         QStandardItem('16:22:11')])
    main_window.active_clients_table.setModel(test_list)
    main_window.active_clients_table.resizeColumnsToContents()
    app.exec_()



"""
В основе метода библиотека dis - анализ кода с помощью его дизассемблирования
(разбор кода на составляющие: в нашем случае - на атрибуты и методы класса)
https://docs.python.org/3/library/dis.html
"""

import dis
from pprint import pprint


class ServerMaker(type):
    """Метакласс для проверки соответствия сервера."""

    def __init__(cls, clsname, bases, clsdict):
        # clsname - экземпляр метакласса - Server
        # bases - кортеж базовых классов - ()
        # clsdict - словарь атрибутов и методов экземпляра метакласса
        # Список методов, которые используются в функциях класса:
        methods = []  # получаем с помощью 'LOAD_GLOBAL'
        # Обычно методы, обёрнутые декораторами попадают
        # не в 'LOAD_GLOBAL', а в 'LOAD_METHOD'
        methods_2 = []  # получаем с помощью 'LOAD_METHOD'
        # Атрибуты, используемые в функциях классов
        attrs = []  # получаем с помощью 'LOAD_ATTR'
        # перебираем ключи
        for func in clsdict:
            try:
                # Возвращает итератор по инструкциям в предоставленной
                # функции, методе, строке исходного кода или объекте
                # кода.
                ret = dis.get_instructions(clsdict[func])
                # Если не функция то ловим исключение
                # (если порт)
            except TypeError:
                pass
            else:
                # Раз функция разбираем код,
                # получая используемые методы и атрибуты.
                for i in ret:
                    # print(i)
                    # opname - имя для операции
                    if i.opname == 'LOAD_GLOBAL':
                        if i.argval not in methods:
                            # заполняем список методами,
                            # использующимися в функциях класса
                            methods.append(i.argval)
                    elif i.opname == 'LOAD_METHOD':
                        if i.argval not in methods_2:
                            # заполняем список атрибутами,
                            # использующимися в функциях класса
                            methods_2.append(i.argval)
                    elif i.opname == 'LOAD_ATTR':
                        if i.argval not in attrs:
                            # заполняем список атрибутами,
                            # использующимися в функциях класса
                            attrs.append(i.argval)
        # print(20*'-', 'methods', 20*'-')
        # pprint(methods)
        # print(20*'-', 'methods_2', 20*'-')
        # pprint(methods_2)
        # print(20*'-', 'attrs', 20*'-')
        # pprint(attrs)
        # print(50*'-')
        # pprint(type(attrs[2]))
        # print(50*'-')
        # Если обнаружено использование недопустимого метода connect,
        # вызываем исключение:
        if 'connect' in methods:
            raise TypeError('Использование метода connect \
                             недопустимо в серверном классе')
        # Если сокет не инициализировался константами
        # SOCK_STREAM(TCP) AF_INET(IPv4), тоже исключение.
        if not ('SOCK_STREAM' in attrs and 'AF_INET' in attrs):
            raise TypeError('Некорректная инициализация сокета.')
        # Обязательно вызываем конструктор предка:
        super().__init__(clsname, bases, clsdict)


# Метакласс для проверки корректности клиентов:
class ClientMaker(type):
    def __init__(cls, clsname, bases, clsdict):
        # Список методов, которые используются в функциях класса:
        methods = []
        for func in clsdict:
            # Пробуем
            try:
                ret = dis.get_instructions(clsdict[func])
                # Если не функция то ловим исключение
            except TypeError:
                pass
            else:
                # Раз функция разбираем код,
                # получая используемые методы.
                for i in ret:
                    if i.opname == 'LOAD_GLOBAL':
                        if i.argval not in methods:
                            methods.append(i.argval)
        pprint(methods)
        # Если обнаружено использование недопустимого метода
        # accept, listen, socket бросаем исключение:
        for command in ('accept', 'listen', 'socket'):
            if command in methods:
                raise TypeError('В классе обнаружено использование \
                                 запрещённого метода')
        # Вызов get_message или send_message из utils
        # считаем корректным использованием сокетов
        if 'get_message' in methods or 'send_message' in methods:
            pass
        else:
            raise TypeError('Отсутствуют вызовы функций, \
                             работающих с сокетами.')
        super().__init__(clsname, bases, clsdict)



"""Декораторы."""
import sys
import socket
import logging
import inspect
sys.path.append('../')


def log(func):
    """Декоратор логирования."""
    def wrapper(*args, **kwargs):
        logger_name = 'server' if 'server.py' in sys.argv[0] \
                      else 'client'
        LOGGER = logging.getLogger(logger_name)

        result = func(*args, **kwargs)

        LOGGER.debug(
            f'Вызвана функция "{func.__name__}" с параметрами \
            args = {args}, kwargs = {kwargs}, ' +
            f'из модуля "{func.__module__}", \
            из функции "{inspect.stack()[1][3]}"')

        return result
    return wrapper


def login_required(func):
    """
    Декоратор, проверяющий, что клиент авторизован на сервере.
    Проверяет, что передаваемый объект сокета находится в
    списке авторизованных клиентов.
    За исключением передачи словаря-запроса
    на авторизацию. Если клиент не авторизован,
    генерирует исключение TypeError
    """

    def checker(*args, **kwargs):
        # проверяем, что первый аргумент - экземпляр MessageProcessor
        # Импортить необходимо тут, иначе ошибка рекурсивного импорта.
        # ------------------------------------------------------------

        from server.core import MyServer

        # if isinstance(args[0], MyServer):
        if True:
            found = False
            print('!!!!!!!!!!! decors login_required')
            for arg in args:
                print('arg =', arg)
                if isinstance(arg, socket.socket):
                    # Проверяем, что данный сокет есть в
                    # списке names класса MyServer
                    for client in args[0].messages:
                        if args[0].messages[client]['socket'] == arg:
                            found = True
                            print('!!!!!!!!!!!!!!! true socket')

            # Теперь надо проверить, что передаваемые аргументы
            # не presence сообщение. Если presence, то разрешаем
            for arg in args:
                if isinstance(arg, dict):
                    if 'action' in arg and arg['action'] == 'presence':
                        found = True
                        print('!!!!!!!!!!!!!!! true presence')
            # Если не не авторизован и не сообщение начала
            # авторизации, то вызываем исключение.
            if not found:
                raise TypeError
        return func(*args, **kwargs)

    return checker



"""Константы."""
import logging


# максимальная длина сообщения в байтах
MAX_PACKAGE_LENGTH = 10240

# кодировка проекта
ENCODING = 'utf-8'

# максимальная очередь подключений
MAX_CONNECTIONS = 5

# параметры подключения к серверу по умолчанию
DEFAULT_ADDRESS = '127.0.0.1'
DEFAULT_PORT = 7777

# Текущий уровень логирования
LOGGING_LEVEL = logging.DEBUG



from setuptools import setup, find_packages

setup(name="ib_message_server",
      version="0.0.3",
      description="Server part of Server to client messanger",
      author="Igor Burakov",
      author_email="iburako8@yandex.ru",
      packages=find_packages(),
      install_requires=['PyQt5', 'sqlalchemy', 'pycryptodome', 'pycryptodomex'],
      scripts=['server/server_run']
      )



"""Функции отправки и получения сообщений через сокет."""
import json
from common.settings import MAX_PACKAGE_LENGTH, ENCODING
from common.decors import log
import common.errors as my_err


@log
def get_message(client):
    """Функция приема и декодирования сообщений."""
    # print(50*'=')
    # print(f'function get_message, client = {client}')
    encoded_response = client.recv(MAX_PACKAGE_LENGTH)
    # print(f'получено сообщение encoded_response = {encoded_response}')
    if isinstance(encoded_response, bytes):
        json_response = encoded_response.decode(ENCODING)
        # print(f'Декодированное сообщение {json_response}')
        if isinstance(json_response, str):
            response = json.loads(json_response)
            if isinstance(response, dict):
                return response
            raise my_err.NonDictError
        raise my_err.NonStrError
    raise my_err.NonBytesError


@log
def send_message(sock, message):
    """Функция кодирования и отправки сообщений."""
    if not isinstance(message, dict):
        raise my_err.NonDictError
    js_message = json.dumps(message)
    encoded_message = js_message.encode(ENCODING)
    sock.send(encoded_message)
    return True



import os
import sys
import unittest
sys.path.insert(0, os.path.join(os.getcwd(), '..'))
from server import process_client_message


class TestProcessClientMessage(unittest.TestCase):
    message = {
            'action': 'presence',
            'time': 111.11,
            'user': {
                'account_name': 'Guest',
            },
        }

    def test_equal(self):
        self.assertEqual(process_client_message(self.message),
                         {'response': 200})
        message_err = self.message
        message_err['action'] = ''  # error message
        self.assertEqual(process_client_message(message_err),
                         {'response': 400,
                          'error': 'bad request'})

    def test_isinstance(self):
        self.assertIsInstance(
            process_client_message(self.message), dict)
        self.assertIsInstance(
            process_client_message(self.message)['response'], int)
        message_err = self.message
        message_err['action'] = ''  # error message
        self.assertIsInstance(
            process_client_message(message_err), dict)
        self.assertIsInstance(
            process_client_message(message_err)['response'], int)
        self.assertIsInstance(
            process_client_message(message_err)['error'], str)


if __name__ == '__main__':
    unittest.main()
